# Generated by Django 4.2.22 on 2025-07-09 10:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0064_alter_collection_unique_together_and_more"),
    ]

    operations = [
        migrations.AddIndex(
            model_name="collectionversion",
            index=models.Index(
                fields=["namespace", "name", "version"], name="ansible_col_namespa_ea6f75_idx"
            ),
        ),
    ]
