# Generated by Django 3.2.16 on 2023-01-17 15:01

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0098_pulp_labels'),
        ('ansible', '0047_ansible_namespace'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionVersionMark',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_collectionversionmark', serialize=False, to='core.content')),
                ('value', models.SlugField()),
                ('marked_collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='marks', to='ansible.collectionversion')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('value', 'marked_collection')},
            },
            bases=('core.content',),
        ),
    ]
