# Generated by Django 2.2.16 on 2020-09-22 20:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0023_alter_requirements_file_field'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='collectionversion',
            name='certification',
        ),
    ]
