# Generated by Django 2.2.6 on 2019-10-18 16:56

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0013_pulp_fields'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='collectionversion',
            name='is_certified',
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='certification',
            field=models.CharField(choices=[('certified', 'certified'), ('not_certified', 'not_certified'), ('needs_review', 'needs_review')], default='needs_review', max_length=13),
        ),
    ]
