# Generated by Django 2.2.13 on 2020-06-09 21:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0018_fix_collection_relative_path'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionremote',
            name='auth_url',
            field=models.CharField(max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='collectionremote',
            name='token',
            field=models.TextField(max_length=2000, null=True),
        ),
    ]
