# Generated by Django 2.2.15 on 2020-08-27 21:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0021_rename_role_remote'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversion',
            name='documentation',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='homepage',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='issues',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='repository',
            field=models.CharField(blank=True, default='', editable=False, max_length=2000),
        ),
    ]
