# Generated by Django 2.2.15 on 2020-08-10 19:26

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0019_collection_token'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='ansiblerepository',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': (('modify_ansible_repo_content', 'Can modify ansible repository content'),)},
        ),
    ]
