# Generated by Django 3.2.18 on 2023-03-01 23:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0048_collectionversionmark'),
    ]

    operations = [
        migrations.AddField(
            model_name='ansiblerepository',
            name='private',
            field=models.BooleanField(default=False),
        ),
        migrations.AlterModelOptions(
            name='ansibledistribution',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_ansibledistribution', 'Can manage roles on distributions')]},
        ),
        migrations.AlterModelOptions(
            name='ansiblerepository',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('rebuild_metadata_ansiblerepository', 'Can rebuild metadata on the repository'), ('repair_ansiblerepository', 'Can repair the repository'), ('sign_ansiblerepository', 'Can sign content on the repository'), ('sync_ansiblerepository', 'Can start a sync task on the repository'), ('manage_roles_ansiblerepository', 'Can manage roles on repositories'), ('modify_ansible_repo_content', 'Can modify repository content')]},
        ),
        migrations.AlterModelOptions(
            name='collectionremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': (('manage_roles_collectionremote', 'Can manage roles on collection remotes'),)},
        ),
        migrations.AlterModelOptions(
            name='gitremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_gitremote', 'Can manage roles on git remotes')]},
        ),
        migrations.AlterModelOptions(
            name='roleremote',
            options={'default_related_name': '%(app_label)s_%(model_name)s', 'permissions': [('manage_roles_roleremote', 'Can manage roles on role remotes')]},
        ),
    ]
