# Generated by Django 4.2.16 on 2024-11-21 08:53

from django.db import migrations
from pulpcore.plugin.migrations import RequireVersion


class Migration(migrations.Migration):

    dependencies = [
        ("ansible", "0059_collectionversion_unique_sha256"),
    ]

    operations = [
        RequireVersion("ansible", "0.23.0"),
        migrations.RemoveConstraint(
            model_name="collectionversion",
            name="unique_is_highest",
        ),
        migrations.RemoveField(
            model_name="collectionversion",
            name="is_highest",
        ),
    ]
