# Generated by Django 4.2.16 on 2024-12-05 12:33

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0037_gitremote'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionVersionSignature',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_collectionversionsignature', serialize=False, to='core.content')),
                ('data', models.BinaryField()),
                ('digest', models.CharField(max_length=64)),
                ('pubkey_fingerprint', models.CharField(max_length=64)),
                ('signed_collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='signatures', to='ansible.collectionversion')),
                ('signing_service', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='signatures', to='core.signingservice')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('pubkey_fingerprint', 'signed_collection')},
            },
            bases=('core.content',),
        ),
    ]
